/***************************************************************************
 *   Copyright (C) 2004-2010 by Stefan Baur                                *
 *   kwatersaver@c0n.de                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <GL/gl.h>
#include "glwater.h"
#include "string.h"

#include "settings.h"

	//We define this for now... as they are experimantal at the moment
//#define NO_GL_EXTENSIONS

#ifndef NO_GL_EXTENSIONS
extern "C" void (*glXGetProcAddressARB(const char *procName))(void);
#endif

#include <sys/time.h>
#include <math.h>

#include <QDebug>

GLWater::GLWater()
	{
	currentTime=0;
	lastWaterStepTime=0;

	Settings.bgTexID=0;

	vbo_vtx=0;
	vbo_idx=0;
	vbo_tc=0;

	water_data=0;
	water_vtx=0;
	water_idx=0;
	water_tc=0;
	water_env_tc=0;
	water_diff=0;
	water_color=0;
	}

GLWater::~GLWater()
	{
	quit();
	}

void GLWater::quit()
	{
	if(water_data)delete[] water_data;water_data=0;
	if(water_vtx)delete[] water_vtx;water_vtx=0;
	if(water_tc)delete[] water_tc;water_tc=0;
	if(water_env_tc)delete[] water_env_tc;water_env_tc=0;
	if(water_diff)delete[] water_diff;water_diff=0;
	if(water_idx)delete[] water_idx;water_idx=0;
	if(water_color)delete[] water_color;water_color=0;
	if(Settings.bgTexID!=0 && glIsTexture(Settings.bgTexID))glDeleteTextures(1,&Settings.bgTexID);
	Settings.bgTexID=0;
	}

bool GLWater::isExtensionSupported(char *str)
	{
#ifdef NO_GL_EXTENSIONS
	str=str;//do nothing and keep compiler happy...
#else
	char *extensions=(char*)glGetString(GL_EXTENSIONS);
	if(glXGetProcAddressARB==0)return false;
	if(strstr(extensions,str))return true;
#endif
	return false;
	}

void GLWater::init()
	{
	int x,y;
	float vtx_stepx,vtx_stepy,vtx_offsx,vtx_offsy,vtx_x,vtx_y;
	INDEX *idx;
	VERTEX *vtx;
	
	qDebug()<<" initializing water:";
	
	qDebug()<<" - OpenGL";

	glDisable(GL_DEPTH_TEST);
	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA,GL_ONE);
	glDisable(GL_BLEND);

	glShadeModel( GL_SMOOTH ); //hey OpenGL, we want some nice rendering :)
	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
	glHint( GL_POINT_SMOOTH_HINT, GL_NICEST );

	qDebug()<<" - OpenGL Extensions";

		// multitexture extension
	if((OpenGLExtensions.ext_multitexture=isExtensionSupported((char*)"GL_ARB_multitexture")))
	{
#ifndef NO_GL_EXTENSIONS
		qDebug()<<"     GL_ARB_multitexture";
		ext.glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC) glXGetProcAddressARB("glActiveTextureARB");
		ext.glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC) glXGetProcAddressARB("glClientActiveTextureARB");
#endif
	}
	else
	{
		ext.glActiveTextureARB = 0;
		ext.glClientActiveTextureARB = 0;
	}
		
		// non power of two texture extension
	if((OpenGLExtensions.ext_NPOT=isExtensionSupported((char*)"GL_ARB_texture_non_power_of_two")))
	{
		qDebug()<<"     GL_ARB_texture_non_power_of_two";
	}
		
		// VBO extension
	if((OpenGLExtensions.ext_VBO=isExtensionSupported((char*)"GL_ARB_vertex_buffer_object")))
		{
#ifndef NO_GL_EXTENSIONS
		qDebug()<<"     GL_ARB_vertex_buffer_object";
		ext.glGenBuffersARB = (PFNGLGENBUFFERSARBPROC) glXGetProcAddressARB("glGenBuffersARB");
		ext.glBindBufferARB = (PFNGLBINDBUFFERARBPROC) glXGetProcAddressARB("glBindBufferARB");
		ext.glBufferDataARB = (PFNGLBUFFERDATAARBPROC) glXGetProcAddressARB("glBufferDataARB");
		ext.glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC) glXGetProcAddressARB("glDeleteBuffersARB");
#endif
		}
	else
		{
		ext.glGenBuffersARB = 0;
		ext.glBindBufferARB = 0;
		ext.glBufferDataARB = 0;
		ext.glDeleteBuffersARB = 0;
		}

	qDebug()<<" - Memory";

		//water effects date storage
	water_data=new WATERTYPE[settings.waterResX*settings.waterResY*2];
	water_vtx=new VERTEX[settings.waterResX*settings.waterResY];
	water_tc=new TEXCOORD[settings.waterResX*settings.waterResY];
	water_env_tc=new TEXCOORD[settings.waterResX*settings.waterResY];
	water_diff=new TEXCOORD[settings.waterResX*settings.waterResY];
	water_idx=new INDEX[(settings.waterResX-1)*(settings.waterResY-1)*4];
	water_color=new COLOR[settings.waterResX*settings.waterResY*3];

	qDebug()<<" - Calm Sea";

		//initialize with zero
	for(y=0;y<settings.waterResY;y++)
		{
		for(x=0;x<settings.waterResX;x++)
			{
			water_data[y*settings.waterResX+x]=0;
			water_data[y*settings.waterResX+x + settings.waterResX*settings.waterResY]=0;

			water_vtx[y*settings.waterResX+x].x=0;water_vtx[y*settings.waterResX+x].y=0;
			water_tc[y*settings.waterResX+x].u=0;water_tc[y*settings.waterResX+x].v=0;
			water_env_tc[y*settings.waterResX+x].u=0;water_env_tc[y*settings.waterResX+x].v=0;
			water_diff[y*settings.waterResX+x].u=0;water_diff[y*settings.waterResX+x].v=0;
			water_color[y*settings.waterResX+x].r=1;
			water_color[y*settings.waterResX+x].g=1;
			water_color[y*settings.waterResX+x].b=1;
			}
		}


	qDebug()<<" - Index Pointering";
	
	idx=water_idx;

		//Create Indices once
	for(y=0;y<settings.waterResY-1;y++)
		{
		for(x=0;x<settings.waterResX-1;x++)
			{
				// (0,0)
			*idx=y*settings.waterResX+x; idx++;
				// (0,1)
			*idx=(y+1)*settings.waterResX+x; idx++;
				// (1,1)
			*idx=(y+1)*settings.waterResX+x+1; idx++;
				// (1,0)
			*idx=y*settings.waterResX+x+1; idx++;
			}
		}

	qDebug()<<" - Texture Coordinates";
	
		//Create Plain TexCoords
	for(y=0;y<settings.waterResY;y++)
		{
		for(x=0;x<settings.waterResX;x++)
			{
			water_tc[y*settings.waterResX+x].u=(float)x/((float)settings.waterResY-1);
			water_tc[y*settings.waterResX+x].v=(float)y/((float)settings.waterResY-1);
		
			water_env_tc[y*settings.waterResX+x].u=0.5+((float)x/((float)settings.waterResY-1)-0.5)/2;
			water_env_tc[y*settings.waterResX+x].v=0.5+((float)y/((float)settings.waterResY-1)-0.5)/2;
			}
		}

	qDebug()<<" - Vertices";
	
		//Create Vertices once
	vtx_offsx=-1;
	vtx_offsy=-1;
	vtx_stepx=2.0f/((float)settings.waterResX-1);
	vtx_stepy=2.0f/((float)settings.waterResY-1);
	vtx=water_vtx;

	for(y=0;y<settings.waterResY;y++)
		{
		vtx_y=((float)y)*vtx_stepy+vtx_offsy;

		for(x=0;x<settings.waterResX;x++)
			{
			vtx_x=((float)x)*vtx_stepx+vtx_offsx;
			vtx->x=vtx_x;vtx->y=vtx_y;vtx++;
			}
		}

	
		//build vertex buffers
	if(OpenGLExtensions.ext_VBO && settings.useVBO)
		{
		qDebug()<<" - VBO";
	
		ext.glGenBuffersARB(1,&vbo_idx);
		ext.glBindBufferARB( GL_ARRAY_BUFFER_ARB, vbo_idx );
		ext.glBufferDataARB( GL_ARRAY_BUFFER_ARB, (settings.waterResX-1)*(settings.waterResY-1)*4*sizeof(INDEX), water_idx, GL_STATIC_DRAW_ARB);

		ext.glGenBuffersARB(1,&vbo_vtx);
		ext.glBindBufferARB( GL_ARRAY_BUFFER_ARB, vbo_vtx );
		ext.glBufferDataARB( GL_ARRAY_BUFFER_ARB, settings.waterResX*settings.waterResY*sizeof(VERTEX), water_vtx, GL_STATIC_DRAW_ARB);

		ext.glGenBuffersARB(1,&vbo_tc);
		ext.glBindBufferARB( GL_ARRAY_BUFFER_ARB, vbo_tc );
		ext.glBufferDataARB( GL_ARRAY_BUFFER_ARB, settings.waterResX*settings.waterResY*sizeof(TEXCOORD), water_tc, GL_STATIC_DRAW_ARB); //GL_DYNAMIC_DRAW_ARB); //FIXME updateable?
		}
		
	qDebug()<<" done.";
	}

bool GLWater::update(double time) //TODO: transparently return the ms, our timer should wait next time
	{
	int x,y;
	float tc_stepx,tc_stepy,tc_u,tc_v,vtx_x,vtx_y,c;
	TEXCOORD *tc,*tc_diff;
	COLOR *col;
	WATERTYPE *w_current;

	currentTime=time;
	
	if(!(lastWaterStepTime<currentTime+1.0/(double)settings.fps)) // we are more than one frame too early -> skip update
	{
		qDebug()<<"(calc-skip) "<<(currentTime-lastWaterStepTime)*settings.fps;
		return false; //no update needed since nothing has to change yet.
	}
	else
	if(!(lastWaterStepTime<currentTime)) // we are just a bit too early, act as if all ok.
	{
		qDebug()<<"( - - - - ) "<<(currentTime-lastWaterStepTime)*settings.fps;
		currentTime+=1.0/(double)settings.fps;
	}

		//calculate each step (without visual update)
	while(lastWaterStepTime<currentTime)
	{
		calcWaterStep();
		lastWaterStepTime+=1.0/(double)settings.fps;
	
		w_current=water_data+(settings.waterResX*settings.waterResY)*((waterStep+1)&1) +settings.waterResX;


		//TODO: Add effects, make them configurable and randomized

			//big psycedelic center blob
		if(settings.effectBlob)
		{
			w_current[settings.waterResX*settings.waterResY/2+settings.waterResX/2]=(int)(10000*10*sin(lastWaterStepTime*13*(double)settings.fps/40)*sin(lastWaterStepTime/21*(double)settings.fps/40)*sin(lastWaterStepTime/21*(double)settings.fps/40));
		}

			//circling blob
		if(settings.effectWhirl)
		{
			vtx_x=0.8f*sin(lastWaterStepTime*(double)settings.fps/30)/2+0.5;
			vtx_y=0.8f*cos(lastWaterStepTime*(double)settings.fps/30)/2+0.5;
			w_current[ (int)( vtx_y*settings.waterResY )*settings.waterResX+(int)( vtx_x*settings.waterResX )]=(int)(10000*10*sin(lastWaterStepTime/3.1415f*(double)settings.fps/40)*sin(lastWaterStepTime/3.1415f*(double)settings.fps/40)*sin(lastWaterStepTime/3.1415f*(double)settings.fps/40)*sin(lastWaterStepTime/3.1415f*(double)settings.fps/40));
		}

		if(settings.effectRain)
		{
			static int rnd;

			rnd+=(int)(settings.waterResX*settings.waterResY*sin(lastWaterStepTime)*sin(lastWaterStepTime));

			rnd%=(settings.waterResX*settings.waterResY);

				//if not on border -> drop rain
			if(rnd&1)if(	rnd>settings.waterResX*2 &&
				rnd<(settings.waterResY-2)*settings.waterResX &&
				rnd%settings.waterResY>1 && rnd%settings.waterResY<settings.waterResY-2 )
			{
				w_current[rnd] = (int)(10000*sin(lastWaterStepTime*345)*10*sin(lastWaterStepTime/15)*sin(lastWaterStepTime/15));
			}
		}
		
		if(lastWaterStepTime<currentTime-1.0/(double)settings.fps) // we still have some more frames to do - we are badly behind
		{
			qDebug()<<"(frameskip) "<<(currentTime-lastWaterStepTime)*settings.fps;
		}
		else if(lastWaterStepTime<currentTime) // we are just a little behind, act as if all is ok.
		{
			qDebug()<<"( - - - - ) "<<(currentTime-lastWaterStepTime)*settings.fps;
		}
	}


		//Ok now everything is updated, make changes visible...

		//calculate water height diffs
	w_current=water_data+(settings.waterResX*settings.waterResY)*((waterStep+1)&1);

	for(y=1;y<settings.waterResY-1;y++)
	{
		for(x=1;x<settings.waterResX-1;x++)
		{
			water_diff[y*settings.waterResX+x].u=(float)(w_current[y*settings.waterResX+x+1]-w_current[y*settings.waterResX+x])*0.00001f;
			water_diff[y*settings.waterResX+x].v=(float)(w_current[(y+1)*settings.waterResX+x]-w_current[y*settings.waterResX+x])*0.00001f;
		}
	}


	tc_stepx=1.0f/((float)settings.waterResX-1);
	tc_stepy=1.0f/((float)settings.waterResY-1);

	//fill texcoords
	tc=water_tc;
	tc_diff=water_diff;

	for(y=0;y<settings.waterResY;y++)
	{
		tc_v=((float)y)*tc_stepy;
		//vtx_y=((float)y)*vtx_stepy+vtx_offsy;

		for(x=0;x<settings.waterResX;x++)
		{
			tc_u=((float)x)*tc_stepx;
			//vtx_x=((float)x)*vtx_stepx+vtx_offsx;

			tc->u=tc_u+tc_diff[0].u;tc->v=tc_v+tc_diff[0].v;tc++;
			//vtx->x=vtx_x;vtx->y=vtx_y;vtx++;

			tc_diff++;
		}
	}

	//fill env texcoords
	if(settings.renderEnvironment)
	{
		tc=water_env_tc;
		tc_diff=water_diff;

		for(y=0;y<settings.waterResY;y++)
		{
			tc_v=(((float)y)*tc_stepy-0.5)/5;

			for(x=0;x<settings.waterResX;x++)
			{
				tc_u=(((float)x)*tc_stepx-0.5)/5;

				tc->u=0.5+tc_u-tc_diff->u*2;tc->v=0.5+tc_v-tc_diff->v*2;tc++;

				tc_diff++;
			}
		}
	}

	//fill color
	if(settings.renderLighting)
	{
		col=water_color;
		tc_diff=water_diff;

		for(y=0;y<settings.waterResY;y++)
		{
			for(x=0;x<settings.waterResX;x++)
			{
					// normal n = (u,v,1)
					// light l = 1/sqrt(3) * (1,1,1)
					// dotprod: n.l = 1/sqrt(3) * u+v+1
				c=(1+tc_diff->u*5-tc_diff->v*5)*0.9;
				col->r=col->g=col->b=c;
	/*			col->r=1-10*(tc_diff->u)*(tc_diff->u);
				col->g=1-10*(tc_diff->v)*(tc_diff->v);
				col->b=1;*/
				col++;
				tc_diff++;
			}
		}
	}


	if(OpenGLExtensions.ext_VBO && settings.useVBO) //FIXME: update texture coords
	{
		ext.glDeleteBuffersARB( 1, &vbo_tc );
		ext.glGenBuffersARB(1,&vbo_tc);
		ext.glBindBufferARB( GL_ARRAY_BUFFER_ARB, vbo_tc );
		ext.glBufferDataARB( GL_ARRAY_BUFFER_ARB, settings.waterResX*settings.waterResY*sizeof(TEXCOORD), water_tc, GL_DYNAMIC_DRAW_ARB);
	}

	return true;
}

void GLWater::render()
{
	//if(0)glClear( GL_COLOR_BUFFER_BIT ); //No need for clearing the screen as we draw over *everything* anyways

	glLoadIdentity();
	  glTranslatef( 0.0,0.0, -50.0f);

	glColor4f(1,1,1,1);

	if(!OpenGLExtensions.ext_multitexture) //single texture
	{
		glEnable(GL_TEXTURE_2D);
		glBindTexture( GL_TEXTURE_2D, Settings.bgTexID);
	}
	else //multi texture
	{
		if(settings.renderEnvironment)
		{
			ext.glActiveTextureARB(GL_TEXTURE0_ARB);
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, Settings.bgTexID);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
			glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, settings.renderLighting?GL_MODULATE:GL_REPLACE);
			
			ext.glActiveTextureARB(GL_TEXTURE1_ARB);
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, Settings.envTexID);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
			glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_MODULATE);
		}
		else
		{
			ext.glActiveTextureARB(GL_TEXTURE0_ARB);
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, Settings.bgTexID);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
			glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, settings.renderLighting?GL_MODULATE:GL_REPLACE);
			glActiveTexture(GL_TEXTURE1);
			glDisable(GL_TEXTURE_2D);
		}
	}

	if(OpenGLExtensions.ext_VBO && settings.useVBO) //use vertex buffers
	{
		glEnableClientState(GL_VERTEX_ARRAY);
		ext.glBindBufferARB(GL_ARRAY_BUFFER_ARB,vbo_vtx);
		glVertexPointer(2,GL_FLOAT,0,0);

		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		ext.glBindBufferARB(GL_ARRAY_BUFFER_ARB,vbo_tc);
		glTexCoordPointer(2,GL_FLOAT,0,0);

		glEnableClientState(GL_INDEX_ARRAY);
		ext.glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbo_idx);

        glDrawElements(GL_QUADS, (settings.waterResX-1)*(settings.waterResY-1)*4, GL_UNSIGNED_SHORT, 0);

		glDisableClientState(GL_INDEX_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
	else	//send the vertex data to the graphics card
	{
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(2,GL_FLOAT,0,water_vtx);
	
		if(settings.renderLighting)
		{
			glEnableClientState(GL_COLOR_ARRAY);
			glColorPointer(3,GL_FLOAT,0,water_color);
		}

		if(!OpenGLExtensions.ext_multitexture) //single texture
		{
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2,GL_FLOAT,0,water_tc);
		}
		else //multi texture
		{
			ext.glClientActiveTextureARB(GL_TEXTURE0_ARB);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2,GL_FLOAT,0,water_tc);

			if(settings.renderEnvironment)
			{
				ext.glClientActiveTextureARB(GL_TEXTURE1_ARB);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(2,GL_FLOAT,0,water_env_tc);
			}
		}
	
		glDrawElements(GL_QUADS,(settings.waterResX-1)*(settings.waterResY-1)*4,GL_UNSIGNED_SHORT,water_idx);

		glDisableClientState(GL_COLOR_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);
		
		if(!OpenGLExtensions.ext_multitexture) //single texture
		{
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}
		else
		{
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			ext.glClientActiveTextureARB(GL_TEXTURE0_ARB);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}
	}

}





void GLWater::calcWaterStep() //simple, nice and relaxing water effect :)
	{
	WATERTYPE val=0;
	int x,y;

	WATERTYPE *w_new=water_data+(settings.waterResX*settings.waterResY)*(waterStep&1)    +settings.waterResX;
	WATERTYPE *w_old=water_data+(settings.waterResX*settings.waterResY)*((waterStep+1)&1) +settings.waterResX;

	for(y=1;y<settings.waterResY-1;y++)
		{
		for(x=0;x<settings.waterResX;x++)
			{
			// (old[-]+old[+]+old+old)/2-current[0]
			val=-w_new[0]+(w_old[-settings.waterResX]+w_old[-1]+w_old[settings.waterResX]+w_old[1])/2;
			*w_new=val-(val>>5);
							//TODO: ideas on how to only update parts of water array to make this even faster?
			w_new++;
			w_old++;
			}
		}

	waterStep++;
	}








