/***************************************************************************
 *   Copyright (C) 2010 by Stefan Baur <kwatersaver@c0n.de>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA            *
 ***************************************************************************/

#include "settings.h"
#include <KConfigGroup>
#include <KGlobal>
#include <KSharedConfig>

#include <QDebug>


void Settings::load()
{
	KConfigGroup config( KGlobal::config(), "Settings" );

	qDebug()<<"Settings::load()";

	fps=config.readEntry("fps",24);
	
	waterResX=config.readEntry("waterResX",128);
	waterResY=config.readEntry("waterResY",128);
	
	bgSource=config.readEntry("bgSource","desktop");
	
	renderLighting=config.readEntry("renderLighting",true);
	renderEnvironment=config.readEntry("renderEnvironment",true);
	
	useVBO=config.readEntry("useVBO",false);
	useMultitexture=config.readEntry("useMultitexture",true);
	usePowerOf2Textures=config.readEntry("usePowerOf2Textures",false);
	
	effectRain=config.readEntry("effectRain",false);
	effectWhirl=config.readEntry("effectWhirl",true);
	effectBlob=config.readEntry("effectBlob",true);
}

void Settings::defaults()
{
	qDebug()<<"Settings::defaults()";

	fps=24;
	
	waterResX=128;
	waterResY=128;
	
	bgSource="desktop";
	
	renderEnvironment=true;
	renderLighting=true;
	
	useVBO=false;
	useMultitexture=true;
	usePowerOf2Textures=false;
	
	effectRain=false;
	effectWhirl=true;
	effectBlob=true;
}

void Settings::save()
{
	KConfigGroup config( KGlobal::config(), "Settings" );

	qDebug()<<"Settings::save()";
	
	config.writeEntry("fps",fps);
	
	config.writeEntry("waterResX",waterResX);
	config.writeEntry("waterResY",waterResY);
	
	config.writeEntry("bgSource",bgSource);
	
	config.writeEntry("renderLighting",renderLighting);
	config.writeEntry("renderEnvironment",renderEnvironment);

	//config.writeEntry("useMultitexture",useMultitexture);
	//config.writeEntry("useVBO",useVBO);
	//config.writeEntry("usePowerOf2Textures",usePowerOf2Textures);

	config.writeEntry("effectRain",effectRain);
	config.writeEntry("effectWhirl",effectWhirl);
	config.writeEntry("effectBlob",effectBlob);

	config.sync();
}

Settings settings;
